from collections import OrderedDict


import rayleaf.stats as stats


class Client:
    
    def __init__(
            self,
            client_num: int,
            client_id: str,
            train_data: dict,
            eval_data: dict,
            model: type,
            model_settings: dict,
            group: list = None,
            device: str = "cpu"
        ) -> None:

        self.client_num = client_num

        self.device = device
        self.model = model(**model_settings)

        self.id = client_id
        self.group = group

        self.train_data = self.model.generate_dataset(train_data)
        self.eval_data = self.model.generate_dataset(eval_data)

        self._num_train_samples = len(self.train_data) if self.train_data is not None else 0
        self._num_eval_samples = len(self.eval_data) if self.eval_data is not None else 0
        self._num_samples = self.num_train_samples + self.num_eval_samples

        self.init()

    
    def init(self) -> None:
        pass


    def train(self):
        self.train_model()

        return self.num_train_samples, self.model_params


    def train_model(self, compute_grads: bool = False):
        if compute_grads:
            self.grads = OrderedDict()
            for param_tensor, layer in self.model_params.items():
                self.grads[param_tensor] = layer.detach().clone()

        self.model = self.model.to(self.device)
        self.model.train_model(self.train_data, self.num_epochs, self.batch_size, self.device)
        self.model = self.model.to("cpu")

        if compute_grads:
            for param_tensor, layer in self.model_params.items():
                self.grads[param_tensor] = self.model_params[param_tensor] - self.grads[param_tensor]


    def _train(self, num_epochs: int = 1, batch_size: int = 10) -> tuple:
        self.num_epochs = num_epochs
        self.batch_size = batch_size

        training_result = self.train()

        return training_result


    def _eval(self, set_to_use: str ="test", batch_size: int = 10) -> dict:
        assert set_to_use in ["train", "test", "val"]

        if set_to_use == "train":
            data = self.train_data
        elif set_to_use == "test" or set_to_use == "val":
            data = self.eval_data

        self.model = self.model.to(self.device)
        eval_metrics = self.model.eval_model(data, batch_size, self.device)
        self.model = self.model.to("cpu")

        eval_metrics[stats.CLIENT_ID_KEY] = self.id
        return eval_metrics


    @property
    def model_params(self) -> OrderedDict:
        return self.model.get_params()


    @model_params.setter
    def model_params(self, params: OrderedDict) -> None:
        self.model.set_params(params)


    @property
    def grads(self) -> OrderedDict:
        return self._grads

    
    @grads.setter
    def grads(self, grads: OrderedDict) -> None:
        self._grads = grads


    @property
    def num_train_samples(self) -> int:
        return self._num_train_samples


    @property
    def num_eval_samples(self) -> int:
        return self._num_eval_samples

    
    @property
    def num_samples(self) -> int:
        return self._num_samples



import rayleaf.utils as utils
import rayleaf.utils.logging_utils as logging_utils

from rayleaf.entities.server import Server
from rayleaf.utils.client_count_utils import client_counts_string, count_selected_clients, online


def train(
    server: Server,
    clients: list,
    clients_per_round: int,
    num_epochs: int,
    batch_size: int,
    round: int,
    num_rounds: int,
    log_progress: bool = True
):
    selected_clients = server.select_clients(round, online(clients), num_clients=clients_per_round)
    if log_progress:
        selected_client_counts = count_selected_clients(selected_clients, clients)
        logging_utils.log(utils.ROUND_STR.format(f"Round {round} of {num_rounds}: Training {clients_per_round} clients: {client_counts_string(selected_client_counts)}"))

    server.train_clients(num_epochs=num_epochs, batch_size=batch_size)
    server._update_model()



from collections import OrderedDict


import torch

from torch import nn, optim
from torch.utils.data import Dataset, DataLoader


import rayleaf.models.utils as model_utils
import rayleaf.stats as stats


class Model(nn.Module):

    def __init__(self, seed: float, lr: float, optimizer=optim.SGD) -> None:
        super(Model, self).__init__()

        self.lr = lr
        self.seed = seed
        self.optimizer = optimizer


    def generate_dataset(self, data: dict) -> Dataset:
        return None


    def train_model(self, data: Dataset, num_epochs: int = 1, batch_size: int = 10, device: str = "cpu") -> None:
        train_dataloader = DataLoader(data, batch_size=batch_size, shuffle=False)

        self.train()
        for _ in range(num_epochs):
            self.run_epoch(train_dataloader, device)


    def run_epoch(self, dataloader: DataLoader, device: str = "cpu") -> None:

        for X, y in dataloader:
            X, y = X.to(device), y.to(device)

            probs = self.forward(X)
            loss = self.loss_fn(probs, y)

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()


    @torch.no_grad()
    def eval_model(self, data: Dataset, batch_size: int = 10, device: str = "cpu") -> dict:
        test_dataloader = DataLoader(data, batch_size=batch_size, shuffle=False)
        size = len(data)
        num_batches = len(test_dataloader)
        self.eval()
        test_loss, correct = 0, 0

        for X, y in test_dataloader:
            X, y = X.to(device), y.to(device)
            
            probs = self.forward(X)
            test_loss += self.loss_fn(probs, y).item()

            preds = model_utils.get_predicted_labels(probs)
            correct += model_utils.number_of_correct(preds, y)

        test_loss /= num_batches

        return {
            stats.NUM_CORRECT_KEY: correct,
            stats.NUM_SAMPLES_KEY: size,
            stats.LOSS_KEY: test_loss
        }


    def get_params(self) -> OrderedDict:
        return self.state_dict()


    def set_params(self, params: OrderedDict) -> None:
        self.load_state_dict(params)



import numpy as np
import os
import tensorflow as tf

from PIL import Image

from rayleaf.models.model import Model


IMAGE_SIZE = 84
IMAGES_DIR = os.path.join('..', 'data', 'celeba', 'data', 'raw', 'img_align_celeba')


class ClientModel(Model):
    def __init__(self, seed, lr, num_classes):
        self.num_classes = num_classes
        super(ClientModel, self).__init__(seed, lr)

    def create_model(self):
        input_ph = tf.placeholder(
            tf.float32, shape=(None, IMAGE_SIZE, IMAGE_SIZE, 3))
        out = input_ph
        for _ in range(4):
            out = tf.layers.conv2d(out, 32, 3, padding='same')
            out = tf.layers.batch_normalization(out, training=True)
            out = tf.layers.max_pooling2d(out, 2, 2, padding='same')
            out = tf.nn.relu(out)
        out = tf.reshape(out, (-1, int(np.prod(out.get_shape()[1:]))))
        logits = tf.layers.dense(out, self.num_classes)
        label_ph = tf.placeholder(tf.int64, shape=(None,))
        loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=label_ph,
            logits=logits)
        predictions = tf.argmax(logits, axis=-1)
        minimize_op = self.optimizer.minimize(
            loss=loss, global_step=tf.train.get_global_step())
        eval_metric_ops = tf.count_nonzero(
            tf.equal(label_ph, tf.argmax(input=logits, axis=1)))
        return input_ph, label_ph, minimize_op, eval_metric_ops, tf.math.reduce_mean(loss)

    def process_x(self, raw_x_batch):
        x_batch = [self._load_image(i) for i in raw_x_batch]
        x_batch = np.array(x_batch)
        return x_batch

    def process_y(self, raw_y_batch):
        return raw_y_batch

    def _load_image(self, img_name):
        img = Image.open(os.path.join(IMAGES_DIR, img_name))
        img = img.resize((IMAGE_SIZE, IMAGE_SIZE)).convert('RGB')
        return np.array(img)



from collections import OrderedDict


import torch
import torchaudio
import torch.nn as nn
import torch.nn.functional as F

from torch.utils.data import Dataset, DataLoader


import rayleaf.models.utils as model_utils

from rayleaf.models.model import Model
from rayleaf.models.speech_commands import utils as sc_utils
import rayleaf.stats as stats


ORIGINAL_FREQ = 16000
RESAMPLE_FREQ = 8000


class ClientModel(Model):
    def __init__(self, seed, lr, num_classes):
        super(ClientModel, self).__init__(seed, lr)

        self.num_classes = num_classes

        self.conv1 = nn.Conv1d(in_channels=1, out_channels=32, kernel_size=80, stride=16)
        self.bn1 = nn.BatchNorm1d(num_features=32)
        self.pool1 = nn.MaxPool1d(kernel_size=4)
        self.conv2 = nn.Conv1d(in_channels=32, out_channels=32, kernel_size=3)
        self.bn2 = nn.BatchNorm1d(num_features=32)
        self.pool2 = nn.MaxPool1d(kernel_size=4)
        self.conv3 = nn.Conv1d(in_channels=32, out_channels=2 * 32, kernel_size=3)
        self.bn3 = nn.BatchNorm1d(num_features=2 * 32)
        self.pool3 = nn.MaxPool1d(kernel_size=4)
        self.conv4 = nn.Conv1d(in_channels=2 * 32, out_channels=2 * 32, kernel_size=3)
        self.bn4 = nn.BatchNorm1d(num_features=2 * 32)
        self.pool4 = nn.MaxPool1d(kernel_size=4)
        self.fc1 = nn.Linear(2 * 32, self.num_classes)

        self.loss_fn = nn.CrossEntropyLoss()
        self.optimizer = self.optimizer(self.parameters(), lr=self.lr)

        self.collate_fn = sc_utils.make_collate_fn(frequency=ORIGINAL_FREQ)

        self.update_running_params = True


    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = F.relu(x)
        x = self.pool1(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = F.relu(x)
        x = self.pool2(x)
        x = self.conv3(x)
        x = self.bn3(x)
        x = F.relu(x)
        x = self.pool3(x)
        x = self.conv4(x)
        x = self.bn4(x)
        x = F.relu(x)
        x = self.pool4(x)
        x = F.avg_pool1d(x, x.shape[-1])
        x = x.permute(0, 2, 1)
        x = self.fc1(x)

        return x


    def train_model(self, data: Dataset, num_epochs: int = 1, batch_size: int = 10, device: str = "cpu") -> None:
        pin_memory = device == "cuda"

        train_dataloader = DataLoader(data, 
            batch_size=batch_size, 
            shuffle=True, 
            drop_last=False,
            collate_fn=self.collate_fn,
            pin_memory=pin_memory
        )

        self.train()
        for _ in range(num_epochs):
            self.run_epoch(train_dataloader, device)


    def run_epoch(self, dataloader: DataLoader, device: str = "cpu") -> None:
        resample = torchaudio.transforms.Resample(orig_freq=ORIGINAL_FREQ, new_freq=RESAMPLE_FREQ).to(device)

        for X, y in dataloader:
            X, y = X.to(device), y.to(device)
            X = resample(X)

            probs = self.forward(X)
            probs = probs.squeeze(dim=1)
            loss = self.loss_fn(probs, y)

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

    
    @torch.no_grad()
    def eval_model(self, data: Dataset, batch_size: int = 10, device: str = "cpu") -> dict:
        pin_memory = device == "cuda"

        test_dataloader = DataLoader(data, 
            batch_size=batch_size, 
            shuffle=False, 
            drop_last=False,
            collate_fn=self.collate_fn,
            pin_memory=pin_memory
        )

        size = len(data)
        num_batches = len(test_dataloader)
        test_loss, correct = 0, 0

        resample = torchaudio.transforms.Resample(orig_freq=ORIGINAL_FREQ, new_freq=RESAMPLE_FREQ).to(device)

        self.eval()
        for X, y in test_dataloader:
            X, y = X.to(device), y.to(device)
            X = resample(X)

            probs = self.forward(X)
            probs = probs.squeeze(dim=1)
            test_loss += self.loss_fn(probs, y).item()

            preds = model_utils.get_predicted_labels(probs)
            correct += model_utils.number_of_correct(preds, y)

        test_loss /= num_batches

        return {
            stats.NUM_CORRECT_KEY: correct,
            stats.NUM_SAMPLES_KEY: size,
            stats.LOSS_KEY: test_loss
        }


    def generate_dataset(self, data: Dataset) -> Dataset:
        return data


    def set_params(self, params: OrderedDict) -> None:
        if self.update_running_params:
            super(ClientModel, self).set_params(params)
        else:
            for param_tensor, layer in params.items():
                if "running_mean" not in param_tensor and "running_var" not in param_tensor:
                    self.state_dict()[param_tensor] = layer.clone().detach()
